r"""Contain dataset generator implementations."""

from __future__ import annotations

__all__ = [
    "BaseDatasetGenerator",
    "VanillaDatasetGenerator",
    "is_dataset_generator_config",
    "setup_dataset_generator",
]

from iden.dataset.generator.base import (
    BaseDatasetGenerator,
    is_dataset_generator_config,
    setup_dataset_generator,
)
from iden.dataset.generator.vanilla import VanillaDatasetGenerator



r"""Contain the base class to implement a data generator."""

from __future__ import annotations

__all__ = ["BaseDataGenerator", "is_data_generator_config", "setup_data_generator"]

import logging
from abc import ABC, abstractmethod
from typing import Generic, TypeVar

from objectory import AbstractFactory
from objectory.utils import is_object_config

T = TypeVar("T")

logger = logging.getLogger(__name__)


class BaseDataGenerator(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Define the base class to generate data.

    Example usage:

    ```pycon
    >>> from iden.data.generator import DataGenerator
    >>> generator = DataGenerator([1, 2, 3])
    >>> generator
    DataGenerator(copy=False)
    >>> generator.generate()
    [1, 2, 3]

    ```
    """

    @abstractmethod
    def generate(self) -> T:
        r"""Generate data.

        Returns:
            The generated data.

        Example usage:

        ```pycon
        >>> from iden.data.generator import DataGenerator
        >>> generator = DataGenerator([1, 2, 3])
        >>> generator.generate()
        [1, 2, 3]

        ```
        """


def is_data_generator_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseDataGenerator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: Specifies the configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration for a
            ``BaseDataGenerator`` object.

    Example usage:

    ```pycon
    >>> from iden.data.generator import is_data_generator_config
    >>> is_data_generator_config({"_target_": "iden.data.generator.DataGenerator"})
    True

    ```
    """
    return is_object_config(config, BaseDataGenerator)


def setup_data_generator(data_generator: BaseDataGenerator | dict) -> BaseDataGenerator:
    r"""Set up a data generator.

    The data generator is instantiated from its configuration by using the
    ``BaseDataGenerator`` factory function.

    Args:
        data_generator: Specifies the data generator or its configuration.

    Returns:
        The instantiated data generator.

    Example usage:

    ```pycon
    >>> from iden.data.generator import is_data_generator_config
    >>> generator = setup_data_generator({"_target_": "iden.data.generator.DataGenerator", "data": [1, 2, 3]})
    >>> generator
    DataGenerator(copy=False)

    ```
    """
    if isinstance(data_generator, dict):
        logger.debug("Initializing a data generator from its configuration...")
        data_generator = BaseDataGenerator.factory(**data_generator)
    if not isinstance(data_generator, BaseDataGenerator):
        logger.warning(
            f"data generator is not a BaseDataGenerator (received: {type(data_generator)})"
        )
    return data_generator



r"""Contain safetensors shard generator implementations."""

from __future__ import annotations

__all__ = ["TorchSafetensorsShardGenerator"]

from typing import TYPE_CHECKING, TypeVar
from unittest.mock import Mock

from coola.utils.imports import check_torch, is_torch_available

from iden.shard import TorchSafetensorsShard, create_torch_safetensors_shard
from iden.shard.generator.file import BaseFileShardGenerator
from iden.utils.imports import check_safetensors

if is_torch_available():
    import torch
else:  # pragma: no cover
    torch = Mock()

if TYPE_CHECKING:
    from pathlib import Path

    from iden.data.generator import BaseDataGenerator

T = TypeVar("T")


class TorchSafetensorsShardGenerator(BaseFileShardGenerator[dict[str, torch.Tensor]]):
    r"""Implement a safetensors shard generator.

    Args:
        data: The data to save in the shard.
        path_uri: The path where to save the URI file.
        path_shard: The path where to save the shard data.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> import torch
    >>> from pathlib import Path
    >>> from iden.data.generator import DataGenerator
    >>> from iden.shard.generator import TorchSafetensorsShardGenerator
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     generator = TorchSafetensorsShardGenerator(
    ...         data=DataGenerator({"key1": torch.ones(2, 3), "key2": torch.arange(5)}),
    ...         path_uri=Path(tmpdir).joinpath("uri"),
    ...         path_shard=Path(tmpdir).joinpath("data"),
    ...     )
    ...     generator
    ...     shard = generator.generate("shard1")
    ...     shard
    ...
    TorchSafetensorsShardGenerator(
      (path_uri): PosixPath('/.../uri')
      (path_shard): PosixPath('/.../data')
      (data): DataGenerator(copy=False)
    )
    TorchSafetensorsShard(uri=file:///.../uri/shard1)

    ```
    """

    def __init__(
        self,
        path_uri: Path,
        path_shard: Path,
        data: BaseDataGenerator[dict[str, torch.Tensor]] | dict,
    ) -> None:
        check_safetensors()
        check_torch()
        super().__init__(path_uri=path_uri, path_shard=path_shard, data=data)

    def _generate(self, data: dict[str, torch.Tensor], shard_id: str) -> TorchSafetensorsShard:
        return create_torch_safetensors_shard(
            data=data,
            uri=self._path_uri.joinpath(shard_id).as_uri(),
            path=self._path_shard.joinpath(shard_id).with_suffix(".safetensors"),
        )



r"""Contain simple data generator implementations."""

from __future__ import annotations

__all__ = ["DataGenerator"]

import copy
from typing import TypeVar

from iden.data.generator.base import BaseDataGenerator

T = TypeVar("T")


class DataGenerator(BaseDataGenerator[T]):
    r"""Implement a simple data generator.

    Args:
        data: The data to return.
        copy: If ``True``, it returns a copy of the data,
            otherwise it always returns the same data.

    Example usage:

    ```pycon
    >>> from iden.data.generator import DataGenerator
    >>> generator = DataGenerator([1, 2, 3])
    >>> generator
    DataGenerator(copy=False)
    >>> generator.generate()
    [1, 2, 3]

    ```
    """

    def __init__(self, data: T, copy: bool = False) -> None:
        self._data = data
        self._copy = bool(copy)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(copy={self._copy})"

    def generate(self) -> T:
        data = self._data
        if self._copy:
            data = copy.deepcopy(data)
        return data



r"""Contain the base class to implement a shard generator."""

from __future__ import annotations

__all__ = ["BaseShardGenerator", "is_shard_generator_config", "setup_shard_generator"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from objectory import AbstractFactory
from objectory.utils import is_object_config

if TYPE_CHECKING:
    from iden.shard import BaseShard

T = TypeVar("T")

logger = logging.getLogger(__name__)


class BaseShardGenerator(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Define the base class to create a shard.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.data.generator import DataGenerator
    >>> from iden.shard.generator import JsonShardGenerator
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     generator = JsonShardGenerator(
    ...         data=DataGenerator([1, 2, 3]),
    ...         path_uri=Path(tmpdir).joinpath("uri"),
    ...         path_shard=Path(tmpdir).joinpath("data"),
    ...     )
    ...     generator
    ...     shard = generator.generate("shard1")
    ...     shard
    ...
    JsonShardGenerator(
      (path_uri): PosixPath('/.../uri')
      (path_shard): PosixPath('/.../data')
      (data): DataGenerator(copy=False)
    )
    JsonShard(uri=file:///.../uri/shard1)

    ```
    """

    @abstractmethod
    def generate(self, shard_id: str) -> BaseShard[T]:
        r"""Generate a shard.

        Args:
            shard_id: The shard IDI.

        Returns:
            The generated shard.

        Example usage:

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.data.generator import DataGenerator
        >>> from iden.shard.generator import JsonShardGenerator
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     generator = JsonShardGenerator(
        ...         data=DataGenerator([1, 2, 3]),
        ...         path_uri=Path(tmpdir).joinpath("uri"),
        ...         path_shard=Path(tmpdir).joinpath("data"),
        ...     )
        ...     shard = generator.generate("shard1")
        ...     shard
        ...
        JsonShard(uri=file:///.../uri/shard1)

        ```
        """


def is_shard_generator_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseShardGenerator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: Specifies the configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration for a
            ``BaseShardGenerator`` object.

    Example usage:

    ```pycon
    >>> from iden.shard.generator import is_shard_generator_config
    >>> is_shard_generator_config({"_target_": "iden.shard.generator.JsonShardGenerator"})
    True

    ```
    """
    return is_object_config(config, BaseShardGenerator)


def setup_shard_generator(shard_generator: BaseShardGenerator | dict) -> BaseShardGenerator:
    r"""Set up a shard generator.

    The shard generator is instantiated from its configuration by using the
    ``BaseShardGenerator`` factory function.

    Args:
        shard_generator: Specifies the shard generator or its configuration.

    Returns:
        The instantiated shard generator.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.shard.generator import setup_shard_generator
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     generator = setup_shard_generator(
    ...         {
    ...             "_target_": "iden.shard.generator.JsonShardGenerator",
    ...             "data": [1, 2, 3],
    ...             "path_uri": Path(tmpdir).joinpath("uri"),
    ...             "path_shard": Path(tmpdir).joinpath("data"),
    ...         }
    ...     )
    ...     generator
    ...
    JsonShardGenerator(
      (path_uri): PosixPath('/.../uri')
      (path_shard): PosixPath('/.../data')
      (data): [1, 2, 3]
    )

    ```
    """
    if isinstance(shard_generator, dict):
        logger.debug("Initializing a shard generator from its configuration...")
        shard_generator = BaseShardGenerator.factory(**shard_generator)
    if not isinstance(shard_generator, BaseShardGenerator):
        logger.warning(
            f"shard generator is not a BaseShardGenerator (received: {type(shard_generator)})"
        )
    return shard_generator



r"""Contain tuple-based shard generator implementations."""

from __future__ import annotations

__all__ = ["ShardTupleGenerator"]

from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping

from iden.shard import BaseShard, ShardTuple, create_shard_tuple
from iden.shard.generator.base import BaseShardGenerator, setup_shard_generator

if TYPE_CHECKING:
    from pathlib import Path


class ShardTupleGenerator(BaseShardGenerator[tuple[BaseShard, ...]]):
    r"""Implement a ``ShardTuple`` generator.

    Args:
        shard: The shard generator or its configuration.
        num_shards: The number of shards to generate in the
            ``ShardTuple``.
        path_uri: The path where to save the URI file.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> import torch
    >>> from pathlib import Path
    >>> from iden.data.generator import DataGenerator
    >>> from iden.shard.generator import ShardTupleGenerator, JsonShardGenerator
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     generator = ShardTupleGenerator(
    ...         shard=JsonShardGenerator(
    ...             data=DataGenerator([1, 2, 3]),
    ...             path_uri=Path(tmpdir).joinpath("uri"),
    ...             path_shard=Path(tmpdir).joinpath("data"),
    ...         ),
    ...         path_uri=Path(tmpdir).joinpath("uri"),
    ...         num_shards=5,
    ...     )
    ...     generator
    ...     shard = generator.generate("shard1")
    ...     shard
    ...
    ShardTupleGenerator(
      (path_uri): PosixPath('/.../uri')
      (num_shards): 5
      (shard): JsonShardGenerator(
          (path_uri): PosixPath('/.../uri')
          (path_shard): PosixPath('/.../data')
          (data): DataGenerator(copy=False)
        )
    )
    ShardTuple(
      (uri): file:///.../uri/shard1
      (shards):
        (0): JsonShard(uri=file:///.../uri/000000001)
        (1): JsonShard(uri=file:///.../uri/000000002)
        (2): JsonShard(uri=file:///.../uri/000000003)
        (3): JsonShard(uri=file:///.../uri/000000004)
        (4): JsonShard(uri=file:///.../uri/000000005)
    )

    ```
    """

    def __init__(self, shard: BaseShardGenerator | dict, num_shards: int, path_uri: Path) -> None:
        self._shard = setup_shard_generator(shard)
        self._num_shards = num_shards
        self._path_uri = path_uri

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "path_uri": self._path_uri,
                    "num_shards": self._num_shards,
                    "shard": self._shard,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, shard_id: str) -> ShardTuple:
        shards = [self._shard.generate(f"{i + 1:09}") for i in range(self._num_shards)]
        return create_shard_tuple(uri=self._path_uri.joinpath(shard_id).as_uri(), shards=shards)



r"""Contain code to load a dataset from its Uniform Resource Identifier
(URI)."""

from __future__ import annotations

__all__ = ["load_from_uri"]

from typing import TYPE_CHECKING

from iden.constants import LOADER
from iden.dataset.loader.base import setup_dataset_loader
from iden.io import load_json
from iden.utils.path import sanitize_path

if TYPE_CHECKING:
    from iden.dataset import BaseDataset


def load_from_uri(uri: str) -> BaseDataset:
    r"""Load a dataset from its Uniform Resource Identifier (URI).

    Args:
        uri: The URI of the dataset.

    Returns:
        The dataset associated to the URI.

    Raises:
        FileNotFoundError: if the URI file does not exist.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.dataset import create_vanilla_dataset, load_from_uri
    >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shards = create_shard_dict(
    ...         shards={
    ...             "train": create_shard_tuple(
    ...                 [
    ...                     create_json_shard(
    ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
    ...                     ),
    ...                     create_json_shard(
    ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...                     ),
    ...                 ],
    ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
    ...             ),
    ...             "val": create_shard_tuple(
    ...                 shards=[],
    ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
    ...             ),
    ...         },
    ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
    ...     )
    ...     assets = create_shard_dict(
    ...         shards={
    ...             "stats": create_json_shard(
    ...                 [1, 2, 3], uri=Path(tmpdir).joinpath("uri_stats").as_uri()
    ...             )
    ...         },
    ...         uri=Path(tmpdir).joinpath("uri_assets").as_uri(),
    ...     )
    ...     uri = Path(tmpdir).joinpath("uri").as_uri()
    ...     _ = create_vanilla_dataset(uri=uri, shards=shards, assets=assets)
    ...     dataset = load_from_uri(uri)
    ...     dataset
    ...
    VanillaDataset(
      (uri): file:///.../uri
      (shards): ShardDict(
          (uri): file:///.../uri_shards
          (shards):
            (train): ShardTuple(
                (uri): file:///.../uri_train
                (shards):
                  (0): JsonShard(uri=file:///.../shard/uri1)
                  (1): JsonShard(uri=file:///.../shard/uri2)
              )
            (val): ShardTuple(
                (uri): file:///.../uri_val
                (shards):
              )
        )
      (assets): ShardDict(
          (uri): file:///.../uri_assets
          (shards):
            (stats): JsonShard(uri=file:///.../uri_stats)
        )
    )

    ```
    """
    path = sanitize_path(uri)
    if not path.is_file():
        msg = f"uri file does not exist: {path}"
        raise FileNotFoundError(msg)
    config = load_json(path)
    loader = setup_dataset_loader(config[LOADER])
    return loader.load(uri)



r"""Contain the base class to implement a dataset generator."""

from __future__ import annotations

__all__ = ["BaseDatasetGenerator", "is_dataset_generator_config", "setup_dataset_generator"]

import logging
from abc import ABC, abstractmethod
from typing import TYPE_CHECKING, Generic, TypeVar

from objectory import AbstractFactory
from objectory.utils import is_object_config

if TYPE_CHECKING:
    from iden.dataset import BaseDataset

T = TypeVar("T")

logger = logging.getLogger(__name__)


class BaseDatasetGenerator(Generic[T], ABC, metaclass=AbstractFactory):
    r"""Define the base class to create a dataset.

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.dataset.generator import VanillaDatasetGenerator
    >>> from iden.shard.generator import ShardDictGenerator
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     generator = VanillaDatasetGenerator(
    ...         path_uri=Path(tmpdir).joinpath("uri"),
    ...         shards=ShardDictGenerator(
    ...             path_uri=Path(tmpdir).joinpath("uri/shards"), shards={}
    ...         ),
    ...         assets=ShardDictGenerator(
    ...             path_uri=Path(tmpdir).joinpath("uri/assets"), shards={}
    ...         ),
    ...     )
    ...     generator
    ...     dataset = generator.generate("dataset1")
    ...     dataset
    ...
    VanillaDatasetGenerator(
      (path_uri): PosixPath('/.../uri')
      (shards): ShardDictGenerator(
          (path_uri): PosixPath('/.../uri/shards')
          (shards):
        )
      (assets): ShardDictGenerator(
          (path_uri): PosixPath('/.../uri/assets')
          (shards):
        )
    )
    VanillaDataset(
      (uri): file:///.../uri/dataset1
      (shards): ShardDict(
          (uri): file:///.../uri/shards/shards
          (shards):
        )
      (assets): ShardDict(
          (uri): file:///.../uri/assets/assets
          (shards):
        )
    )

    ```
    """

    @abstractmethod
    def generate(self, dataset_id: str) -> BaseDataset[T]:
        r"""Generate a dataset.

        Args:
            dataset_id: The dataset IDI.

        Returns:
            The generated dataset.

        ```pycon
        >>> import tempfile
        >>> from pathlib import Path
        >>> from iden.dataset.generator import VanillaDatasetGenerator
        >>> from iden.shard.generator import ShardDictGenerator
        >>> with tempfile.TemporaryDirectory() as tmpdir:
        ...     generator = VanillaDatasetGenerator(
        ...         path_uri=Path(tmpdir).joinpath("uri"),
        ...         shards=ShardDictGenerator(
        ...             path_uri=Path(tmpdir).joinpath("uri/shards"), shards={}
        ...         ),
        ...         assets=ShardDictGenerator(
        ...             path_uri=Path(tmpdir).joinpath("uri/assets"), shards={}
        ...         ),
        ...     )
        ...     dataset = generator.generate("dataset1")
        ...     dataset
        ...
        VanillaDataset(
          (uri): file:///.../uri/dataset1
          (shards): ShardDict(
              (uri): file:///.../uri/shards/shards
              (shards):
            )
          (assets): ShardDict(
              (uri): file:///.../uri/assets/assets
              (shards):
            )
        )

        ```
        """


def is_dataset_generator_config(config: dict) -> bool:
    r"""Indicate if the input configuration is a configuration for a
    ``BaseDatasetGenerator``.

    This function only checks if the value of the key  ``_target_``
    is valid. It does not check the other values. If ``_target_``
    indicates a function, the returned type hint is used to check
    the class.

    Args:
        config: Specifies the configuration to check.

    Returns:
        ``True`` if the input configuration is a configuration for a
            ``BaseDatasetGenerator`` object.

    Example usage:

    ```pycon
    >>> from iden.dataset.generator import is_dataset_generator_config
    >>> is_dataset_generator_config(
    ...     {"_target_": "iden.dataset.generator.VanillaDatasetGenerator"}
    ... )
    True

    ```
    """
    return is_object_config(config, BaseDatasetGenerator)


def setup_dataset_generator(dataset_generator: BaseDatasetGenerator | dict) -> BaseDatasetGenerator:
    r"""Set up a dataset generator.

    The dataset generator is instantiated from its configuration by using the
    ``BaseDatasetGenerator`` factory function.

    Args:
        dataset_generator: Specifies the dataset generator or its configuration.

    Returns:
        The instantiated dataset generator.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.dataset.generator import setup_dataset_generator
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     generator = setup_dataset_generator(
    ...         {
    ...             "_target_": "iden.dataset.generator.VanillaDatasetGenerator",
    ...             "path_uri": Path(tmpdir).joinpath("uri"),
    ...             "shards": {
    ...                 "_target_": "iden.shard.generator.ShardDictGenerator",
    ...                 "path_uri": Path(tmpdir).joinpath("uri/shards"),
    ...                 "shards": {}
    ...             },
    ...             "assets": {
    ...                 "_target_": "iden.shard.generator.ShardDictGenerator",
    ...                 "path_uri": Path(tmpdir).joinpath("uri/assets"),
    ...                 "shards": {}
    ...             }
    ...         }
    ...     )
    ...     generator
    ...
    VanillaDatasetGenerator(
      (path_uri): PosixPath('/.../uri')
      (shards): ShardDictGenerator(
          (path_uri): PosixPath('/.../uri/shards')
          (shards):
        )
      (assets): ShardDictGenerator(
          (path_uri): PosixPath('/.../uri/assets')
          (shards):
        )
    )

    ```
    """
    if isinstance(dataset_generator, dict):
        logger.debug("Initializing a dataset generator from its configuration...")
        dataset_generator = BaseDatasetGenerator.factory(**dataset_generator)
    if not isinstance(dataset_generator, BaseDatasetGenerator):
        logger.warning(
            f"dataset generator is not a BaseDatasetGenerator (received: {type(dataset_generator)})"
        )
    return dataset_generator



r"""Contain torch shard generator implementations."""

from __future__ import annotations

__all__ = ["TorchShardGenerator"]

from typing import TYPE_CHECKING, TypeVar

from coola.utils.imports import check_torch

from iden.shard import TorchShard, create_torch_shard
from iden.shard.generator.file import BaseFileShardGenerator

if TYPE_CHECKING:
    from pathlib import Path

    from iden.data.generator import BaseDataGenerator

T = TypeVar("T")


class TorchShardGenerator(BaseFileShardGenerator[T]):
    r"""Implement a torch shard generator.

    Args:
        data: The data to save in the shard.
        path_uri: The path where to save the URI file.
        path_shard: The path where to save the shard data.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.data.generator import DataGenerator
    >>> from iden.shard.generator import TorchShardGenerator
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     generator = TorchShardGenerator(
    ...         data=DataGenerator([1, 2, 3]),
    ...         path_uri=Path(tmpdir).joinpath("uri"),
    ...         path_shard=Path(tmpdir).joinpath("data"),
    ...     )
    ...     generator
    ...     shard = generator.generate("shard1")
    ...     shard
    ...
    TorchShardGenerator(
      (path_uri): PosixPath('/.../uri')
      (path_shard): PosixPath('/.../data')
      (data): DataGenerator(copy=False)
    )
    TorchShard(uri=file:///.../uri/shard1)

    ```
    """

    def __init__(self, data: BaseDataGenerator[T] | dict, path_uri: Path, path_shard: Path) -> None:
        check_torch()
        super().__init__(data=data, path_uri=path_uri, path_shard=path_shard)

    def _generate(self, data: T, shard_id: str) -> TorchShard[T]:
        return create_torch_shard(
            data=data,
            uri=self._path_uri.joinpath(shard_id).as_uri(),
            path=self._path_shard.joinpath(shard_id).with_suffix(".pt"),
        )



r"""Contain I/O utility functions."""

from __future__ import annotations

__all__ = ["generate_unique_tmp_path", "get_loader_mapping"]

import uuid
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from pathlib import Path

    from iden.io import BaseLoader


def generate_unique_tmp_path(path: Path) -> Path:
    r"""Return a unique temporary path given a path.

    This function updates the name to add a UUID.

    Args:
        path: The input path.

    Returns:
        The unique name.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.io.utils import generate_unique_tmp_path
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = generate_unique_tmp_path(Path(tmpdir).joinpath("data.pt"))
    ...     path
    ...
    PosixPath('/.../data-....pt')

    ```
    """
    h = uuid.uuid4().hex
    extension = "".join(path.suffixes)[1:]
    if extension:
        extension = "." + extension
        stem = path.name[: -len(extension)]
    else:
        stem = path.name
    return path.with_name(f"{stem}-{h}{extension}")


def get_loader_mapping() -> dict[str, BaseLoader]:
    r"""Get a default mapping between the file extensions and loaders.

    Returns:
        The mapping between the file extensions and loaders.

    Example usage:

    ```pycon
    >>> from iden.io.utils import get_loader_mapping
    >>> get_loader_mapping()
    {'json': JsonLoader(), 'pkl': PickleLoader(), 'pickle': PickleLoader(), ...}

    ```
    """
    from iden import io  # Local import to avoid cyclic dependencies

    return (
        io.json.get_loader_mapping()
        | io.pickle.get_loader_mapping()
        | io.text.get_loader_mapping()
        | io.torch.get_loader_mapping()
        | io.yaml.get_loader_mapping()
    )



r"""Contain data generator implementations."""

from __future__ import annotations

__all__ = [
    "BaseDataGenerator",
    "DataGenerator",
    "is_data_generator_config",
    "setup_data_generator",
]

from iden.data.generator.base import (
    BaseDataGenerator,
    is_data_generator_config,
    setup_data_generator,
)
from iden.data.generator.vanilla import DataGenerator



r"""Contain ``VanillaDataset`` generator implementations."""

from __future__ import annotations

__all__ = ["VanillaDatasetGenerator"]

from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping, str_indent, str_mapping

from iden.dataset import VanillaDataset, create_vanilla_dataset
from iden.dataset.generator import BaseDatasetGenerator
from iden.shard import BaseShard
from iden.shard.generator.base import setup_shard_generator

if TYPE_CHECKING:
    from pathlib import Path

    from iden.shard.generator import ShardDictGenerator


class VanillaDatasetGenerator(BaseDatasetGenerator[tuple[BaseShard, ...]]):
    r"""Implement a ``VanillaDataset`` generator.

    Args:
        path_uri: The path where to save the URI file.
        shards: The shards generator or its configuration.
        assets: The assets generator or its configuration.

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.dataset.generator import VanillaDatasetGenerator
    >>> from iden.shard.generator import ShardDictGenerator
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     generator = VanillaDatasetGenerator(
    ...         path_uri=Path(tmpdir).joinpath("uri"),
    ...         shards=ShardDictGenerator(
    ...             path_uri=Path(tmpdir).joinpath("uri/shards"), shards={}
    ...         ),
    ...         assets=ShardDictGenerator(
    ...             path_uri=Path(tmpdir).joinpath("uri/assets"), shards={}
    ...         ),
    ...     )
    ...     generator
    ...     dataset = generator.generate("dataset1")
    ...     dataset
    ...
    VanillaDatasetGenerator(
      (path_uri): PosixPath('/.../uri')
      (shards): ShardDictGenerator(
          (path_uri): PosixPath('/.../uri/shards')
          (shards):
        )
      (assets): ShardDictGenerator(
          (path_uri): PosixPath('/.../uri/assets')
          (shards):
        )
    )
    VanillaDataset(
      (uri): file:///.../uri/dataset1
      (shards): ShardDict(
          (uri): file:///.../uri/shards/shards
          (shards):
        )
      (assets): ShardDict(
          (uri): file:///.../uri/assets/assets
          (shards):
        )
    )

    ```
    """

    def __init__(
        self,
        path_uri: Path,
        shards: ShardDictGenerator | dict,
        assets: ShardDictGenerator | dict,
    ) -> None:
        self._path_uri = path_uri
        self._shards = setup_shard_generator(shards)
        self._assets = setup_shard_generator(assets)

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "path_uri": self._path_uri,
                    "shards": self._shards,
                    "assets": self._assets,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def __str__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "path_uri": self._path_uri,
                    "shards": self._shards,
                    "assets": self._assets,
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, dataset_id: str) -> VanillaDataset:
        shards = self._shards.generate(shard_id="shards")
        assets = self._assets.generate(shard_id="assets")
        return create_vanilla_dataset(
            uri=self._path_uri.joinpath(dataset_id).as_uri(),
            shards=shards,
            assets=assets,
        )



r"""Contain text-based data loaders and savers."""

from __future__ import annotations

__all__ = ["TextLoader", "TextSaver", "load_text", "save_text", "get_loader_mapping"]

from pathlib import Path
from typing import Any, TypeVar

from iden.io.base import BaseFileSaver, BaseLoader

T = TypeVar("T")


class TextLoader(BaseLoader[Any]):
    r"""Implement a data loader to load data in a text file.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.io import save_text, TextLoader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("data.txt")
    ...     save_text("hello", path)
    ...     data = TextLoader().load(path)
    ...     data
    ...
    'hello'

    ```
    """

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def load(self, path: Path) -> Any:
        with Path.open(path) as file:
            return file.read()


class TextSaver(BaseFileSaver[Any]):
    r"""Implement a file saver to save data with a text file.

    Note:
        If the data to save is not a string, it is converted to
            a string before to be saved by using ``str``.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.io import TextSaver, TextLoader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("data.txt")
    ...     TextSaver().save("hello", path)
    ...     data = TextLoader().load(path)
    ...     data
    ...
    'hello'

    ```
    """

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def _save_file(self, to_save: Any, path: Path) -> None:
        with Path.open(path, mode="w") as file:
            file.write(str(to_save))


def load_text(path: Path) -> str:
    r"""Load the data from a given text file.

    Args:
        path: Specifies the path where to the text file.

    Returns:
        The data from the text file.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.io import save_text, load_text
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("data.txt")
    ...     save_text("hello", path)
    ...     data = load_text(path)
    ...     data
    ...
    'hello'

    ```
    """
    return TextLoader().load(path)


def save_text(to_save: Any, path: Path, *, exist_ok: bool = False) -> None:
    r"""Save the given data in a text file.

    Args:
        to_save: Specifies the data to write in a text file.
        path: Specifies the path where to write the text file.
        exist_ok: If ``exist_ok`` is ``False`` (the default),
            ``FileExistsError`` is raised if the target file
            already exists. If ``exist_ok`` is ``True``,
            ``FileExistsError`` will not be raised unless the
            given path already exists in the file system and is
            not a file.

    Raises:
        FileExistsError: if the file already exists.

    Note:
        If the data to save is not a string, it is converted to
            a string before to be saved by using ``str``.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.io import save_text, load_text
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("data.txt")
    ...     save_text("hello", path)
    ...     data = load_text(path)
    ...     data
    ...
    'hello'

    ```
    """
    TextSaver().save(to_save, path, exist_ok=exist_ok)


def get_loader_mapping() -> dict[str, BaseLoader]:
    r"""Get a default mapping between the file extensions and loaders.

    Returns:
        The mapping between the file extensions and loaders.

    Example usage:

    ```pycon
    >>> from iden.io.text import get_loader_mapping
    >>> get_loader_mapping()
    {'txt': TextLoader()}

    ```
    """
    return {"txt": TextLoader()}



r"""Contain pickle-based data loaders and savers."""

from __future__ import annotations

__all__ = ["PickleLoader", "PickleSaver", "load_pickle", "save_pickle", "get_loader_mapping"]

import pickle
from pathlib import Path
from typing import Any, TypeVar

from iden.io.base import BaseFileSaver, BaseLoader

T = TypeVar("T")


class PickleLoader(BaseLoader[Any]):
    r"""Implement a data loader to load data in a pickle file.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.io import save_pickle, PickleLoader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("data.pkl")
    ...     save_pickle({"key1": [1, 2, 3], "key2": "abc"}, path)
    ...     data = PickleLoader().load(path)
    ...     data
    ...
    {'key1': [1, 2, 3], 'key2': 'abc'}

    ```
    """

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def load(self, path: Path) -> Any:
        with Path.open(path, mode="rb") as file:
            return pickle.load(file)  # noqa: S301


class PickleSaver(BaseFileSaver[Any]):
    r"""Implement a file saver to save data with a pickle file.

    Args:
        protocol: The pickle protocol. By default, it uses the
            highest protocol available.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.io import PickleSaver, PickleLoader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("data.pkl")
    ...     PickleSaver().save({"key1": [1, 2, 3], "key2": "abc"}, path)
    ...     data = PickleLoader().load(path)
    ...     data
    ...
    {'key1': [1, 2, 3], 'key2': 'abc'}

    ```
    """

    def __init__(self, protocol: int = pickle.HIGHEST_PROTOCOL) -> None:
        self._protocol = protocol

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}(protocol={self._protocol})"

    def _save_file(self, to_save: Any, path: Path) -> None:
        with Path.open(path, mode="wb") as file:
            pickle.dump(to_save, file, protocol=self._protocol)


def load_pickle(path: Path) -> Any:
    r"""Load the data from a given pickle file.

    Args:
        path: Specifies the path to the pickle file.

    Returns:
        The data from the pickle file.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.io import save_pickle, load_pickle
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("data.pkl")
    ...     save_pickle({"key1": [1, 2, 3], "key2": "abc"}, path)
    ...     data = load_pickle(path)
    ...     data
    ...
    {'key1': [1, 2, 3], 'key2': 'abc'}

    ```
    """
    return PickleLoader().load(path)


def save_pickle(
    to_save: Any, path: Path, *, exist_ok: bool = False, protocol: int = pickle.HIGHEST_PROTOCOL
) -> None:
    r"""Save the given data in a pickle file.

    Args:
        to_save: Specifies the data to write in a pickle file.
        path: Specifies the path where to write the pickle file.
        exist_ok: If ``exist_ok`` is ``False`` (the default),
            ``FileExistsError`` is raised if the target file
            already exists. If ``exist_ok`` is ``True``,
            ``FileExistsError`` will not be raised unless the
            given path already exists in the file system and is
            not a file.
        protocol: Specifies the pickle protocol. By default,
            it uses the highest protocol available.

    Raises:
        FileExistsError: if the file already exists.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.io import save_pickle, load_pickle
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("data.pkl")
    ...     save_pickle({"key1": [1, 2, 3], "key2": "abc"}, path)
    ...     data = load_pickle(path)
    ...     data
    ...
    {'key1': [1, 2, 3], 'key2': 'abc'}

    ```
    """
    PickleSaver(protocol=protocol).save(to_save, path, exist_ok=exist_ok)


def get_loader_mapping() -> dict[str, BaseLoader]:
    r"""Get a default mapping between the file extensions and loaders.

    Returns:
        The mapping between the file extensions and loaders.

    Example usage:

    ```pycon
    >>> from iden.io.pickle import get_loader_mapping
    >>> get_loader_mapping()
    {'pkl': PickleLoader(), 'pickle': PickleLoader()}

    ```
    """
    loader = PickleLoader()
    return {"pkl": loader, "pickle": loader}



r"""Contain shard loader implementations for
``VanillaDatasetLoader``."""

from __future__ import annotations

__all__ = ["VanillaDatasetLoader"]

from typing import TypeVar

from iden.dataset.loader.base import BaseDatasetLoader
from iden.dataset.vanilla import VanillaDataset

T = TypeVar("T")


class VanillaDatasetLoader(BaseDatasetLoader[T]):
    r"""Implement a ``VanillaDatasetLoader`` loader.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.dataset import create_vanilla_dataset
    >>> from iden.dataset.loader import VanillaDatasetLoader
    >>> from iden.shard import create_json_shard, create_shard_dict, create_shard_tuple
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     shards = create_shard_dict(
    ...         shards={
    ...             "train": create_shard_tuple(
    ...                 [
    ...                     create_json_shard(
    ...                         [1, 2, 3], uri=Path(tmpdir).joinpath("shard/uri1").as_uri()
    ...                     ),
    ...                     create_json_shard(
    ...                         [4, 5, 6, 7], uri=Path(tmpdir).joinpath("shard/uri2").as_uri()
    ...                     ),
    ...                 ],
    ...                 uri=Path(tmpdir).joinpath("uri_train").as_uri(),
    ...             ),
    ...             "val": create_shard_tuple(
    ...                 shards=[],
    ...                 uri=Path(tmpdir).joinpath("uri_val").as_uri(),
    ...             ),
    ...         },
    ...         uri=Path(tmpdir).joinpath("uri_shards").as_uri(),
    ...     )
    ...     assets = create_shard_dict(
    ...         shards={
    ...             "stats": create_json_shard(
    ...                 [1, 2, 3], uri=Path(tmpdir).joinpath("uri_stats").as_uri()
    ...             )
    ...         },
    ...         uri=Path(tmpdir).joinpath("uri_assets").as_uri(),
    ...     )
    ...     uri = Path(tmpdir).joinpath("uri").as_uri()
    ...     _ = create_vanilla_dataset(uri=uri, shards=shards, assets=assets)
    ...     loader = VanillaDatasetLoader()
    ...     dataset = loader.load(uri)
    ...     dataset
    ...
    VanillaDataset(
      (uri): file:///.../uri
      (shards): ShardDict(
          (uri): file:///.../uri_shards
          (shards):
            (train): ShardTuple(
                (uri): file:///.../uri_train
                (shards):
                  (0): JsonShard(uri=file:///.../shard/uri1)
                  (1): JsonShard(uri=file:///.../shard/uri2)
              )
            (val): ShardTuple(
                (uri): file:///.../uri_val
                (shards):
              )
        )
      (assets): ShardDict(
          (uri): file:///.../uri_assets
          (shards):
            (stats): JsonShard(uri=file:///.../uri_stats)
        )
    )

    ```
    """

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def load(self, uri: str) -> VanillaDataset[T]:
        return VanillaDataset.from_uri(uri)



r"""Contain dict-based shard generator implementations."""

from __future__ import annotations

__all__ = ["ShardDictGenerator"]

from typing import TYPE_CHECKING

from coola.utils import repr_indent, repr_mapping, str_indent, str_mapping

from iden.shard import BaseShard, ShardDict, create_shard_dict
from iden.shard.generator.base import BaseShardGenerator, setup_shard_generator

if TYPE_CHECKING:
    from pathlib import Path


class ShardDictGenerator(BaseShardGenerator[tuple[BaseShard, ...]]):
    r"""Implement a ``ShardDict`` generator.

    Args:
        shards: The shard generators or their configurations.
        path_uri: The path where to save the URI file.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> import torch
    >>> from pathlib import Path
    >>> from iden.data.generator import DataGenerator
    >>> from iden.shard.generator import ShardDictGenerator, JsonShardGenerator
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     generator = ShardDictGenerator(
    ...         shards={
    ...             "train": JsonShardGenerator(
    ...                 data=DataGenerator([1, 2, 3]),
    ...                 path_uri=Path(tmpdir).joinpath("uri"),
    ...                 path_shard=Path(tmpdir).joinpath("data"),
    ...             )
    ...         },
    ...         path_uri=Path(tmpdir).joinpath("uri"),
    ...     )
    ...     generator
    ...     shard = generator.generate("shard1")
    ...     shard
    ...
    ShardDictGenerator(
      (path_uri): PosixPath('/.../uri')
      (shards):
        (train): JsonShardGenerator(
            (path_uri): PosixPath('/.../uri')
            (path_shard): PosixPath('/.../data')
            (data): DataGenerator(copy=False)
          )
    )
    ShardDict(
      (uri): file:///.../uri/shard1
      (shards):
        (train): JsonShard(uri=file:///.../uri/train)
    )

    ```
    """

    def __init__(self, shards: dict[str, BaseShardGenerator | dict], path_uri: Path) -> None:
        self._shards = {key: setup_shard_generator(shard) for key, shard in shards.items()}
        self._path_uri = path_uri

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {
                    "path_uri": self._path_uri,
                    "shards": "\n" + repr_mapping(self._shards),
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def __str__(self) -> str:
        args = str_indent(
            str_mapping(
                {
                    "path_uri": self._path_uri,
                    "shards": "\n" + str_mapping(self._shards),
                }
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, shard_id: str) -> ShardDict:
        shards = {key: shard.generate(str(key)) for key, shard in self._shards.items()}
        return create_shard_dict(uri=self._path_uri.joinpath(shard_id).as_uri(), shards=shards)



r"""Contain shard generator implementations."""

from __future__ import annotations

__all__ = [
    "BaseShardGenerator",
    "JsonShardGenerator",
    "PickleShardGenerator",
    "ShardDictGenerator",
    "ShardTupleGenerator",
    "TorchSafetensorsShardGenerator",
    "TorchShardGenerator",
    "YamlShardGenerator",
    "is_shard_generator_config",
    "setup_shard_generator",
]

from iden.shard.generator.base import (
    BaseShardGenerator,
    is_shard_generator_config,
    setup_shard_generator,
)
from iden.shard.generator.dict import ShardDictGenerator
from iden.shard.generator.json import JsonShardGenerator
from iden.shard.generator.pickle import PickleShardGenerator
from iden.shard.generator.safetensors import TorchSafetensorsShardGenerator
from iden.shard.generator.torch import TorchShardGenerator
from iden.shard.generator.tuple import ShardTupleGenerator
from iden.shard.generator.yaml import YamlShardGenerator



r"""Contain torch-based data loaders and savers."""

from __future__ import annotations

__all__ = ["TorchLoader", "TorchSaver", "load_torch", "save_torch", "get_loader_mapping"]

from typing import TYPE_CHECKING, Any
from unittest.mock import Mock

from coola.utils import check_torch, is_torch_available

from iden.io.base import BaseFileSaver, BaseLoader

if is_torch_available():
    import torch
else:  # pragma: no cover
    torch = Mock()

if TYPE_CHECKING:
    from pathlib import Path


class TorchLoader(BaseLoader[Any]):
    r"""Implement a data loader to load data in a JSON file.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.io import save_torch, TorchLoader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("data.pt")
    ...     save_torch({"key1": [1, 2, 3], "key2": "abc"}, path)
    ...     data = TorchLoader().load(path)
    ...     data
    ...
    {'key1': [1, 2, 3], 'key2': 'abc'}

    ```
    """

    def __init__(self) -> None:
        check_torch()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def load(self, path: Path) -> Any:
        return torch.load(path)


class TorchSaver(BaseFileSaver[Any]):
    r"""Implement a file saver to save data with a JSON file.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.io import TorchSaver, TorchLoader
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("data.pt")
    ...     TorchSaver().save({"key1": [1, 2, 3], "key2": "abc"}, path)
    ...     data = TorchLoader().load(path)
    ...     data
    ...
    {'key1': [1, 2, 3], 'key2': 'abc'}

    ```
    """

    def __init__(self) -> None:
        check_torch()

    def __eq__(self, other: object) -> bool:
        return isinstance(other, self.__class__)

    def __repr__(self) -> str:
        return f"{self.__class__.__qualname__}()"

    def _save_file(self, to_save: Any, path: Path) -> None:
        torch.save(to_save, path)


def load_torch(path: Path) -> Any:
    r"""Load the data from a given JSON file.

    Args:
        path: Specifies the path to the JSON file.

    Returns:
        The data from the JSON file.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.io import save_torch, load_torch
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("data.pt")
    ...     save_torch({"key1": [1, 2, 3], "key2": "abc"}, path)
    ...     data = load_torch(path)
    ...     data
    ...
    {'key1': [1, 2, 3], 'key2': 'abc'}

    ```
    """
    return TorchLoader().load(path)


def save_torch(to_save: Any, path: Path, *, exist_ok: bool = False) -> None:
    r"""Save the given data in a JSON file.

    Args:
        to_save: Specifies the data to write in a JSON file.
        path: Specifies the path where to write the JSON file.
        exist_ok: If ``exist_ok`` is ``False`` (the default),
            ``FileExistsError`` is raised if the target file
            already exists. If ``exist_ok`` is ``True``,
            ``FileExistsError`` will not be raised unless the
            given path already exists in the file system and is
            not a file.

    Raises:
        FileExistsError: if the file already exists.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.io import save_torch, load_torch
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     path = Path(tmpdir).joinpath("data.pt")
    ...     save_torch({"key1": [1, 2, 3], "key2": "abc"}, path)
    ...     data = load_torch(path)
    ...     data
    ...
    {'key1': [1, 2, 3], 'key2': 'abc'}

    ```
    """
    TorchSaver().save(to_save, path, exist_ok=exist_ok)


def get_loader_mapping() -> dict[str, BaseLoader]:
    r"""Get a default mapping between the file extensions and loaders.

    Returns:
        The mapping between the file extensions and loaders.

    Example usage:

    ```pycon
    >>> from iden.io.torch import get_loader_mapping
    >>> get_loader_mapping()
    {'pt': TorchLoader()}

    ```
    """
    if not is_torch_available():
        return {}
    return {"pt": TorchLoader()}



r"""Contain file-based shard generator implementations."""

from __future__ import annotations

__all__ = ["BaseFileShardGenerator"]

from abc import abstractmethod
from typing import TYPE_CHECKING, TypeVar

from coola.utils import repr_indent, repr_mapping, str_indent, str_mapping

from iden.data.generator import BaseDataGenerator, setup_data_generator
from iden.shard.generator.base import BaseShardGenerator

if TYPE_CHECKING:
    from pathlib import Path

    from iden.shard import JsonShard

T = TypeVar("T")


class BaseFileShardGenerator(BaseShardGenerator[T]):
    r"""Implement a JSON shard generator.

    Args:
        data: The data generator or its configuration.
        path_uri: The path where to save the URI file.
        path_shard: The path where to save the shard data.

    Example usage:

    ```pycon
    >>> import tempfile
    >>> from pathlib import Path
    >>> from iden.data.generator import DataGenerator
    >>> from iden.shard.generator import JsonShardGenerator
    >>> with tempfile.TemporaryDirectory() as tmpdir:
    ...     generator = JsonShardGenerator(
    ...         data=DataGenerator([1, 2, 3]),
    ...         path_uri=Path(tmpdir).joinpath("uri"),
    ...         path_shard=Path(tmpdir).joinpath("data"),
    ...     )
    ...     generator
    ...     shard = generator.generate("shard1")
    ...     shard
    ...
    JsonShardGenerator(
      (path_uri): PosixPath('/.../uri')
      (path_shard): PosixPath('/.../data')
      (data): DataGenerator(copy=False)
    )
    JsonShard(uri=file:///.../uri/shard1)

    ```
    """

    def __init__(self, path_uri: Path, path_shard: Path, data: BaseDataGenerator[T] | dict) -> None:
        self._data = setup_data_generator(data)
        self._path_uri = path_uri
        self._path_shard = path_shard

    def __repr__(self) -> str:
        args = repr_indent(
            repr_mapping(
                {"path_uri": self._path_uri, "path_shard": self._path_shard, "data": self._data}
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def __str__(self) -> str:
        args = str_indent(
            str_mapping(
                {"path_uri": self._path_uri, "path_shard": self._path_shard, "data": self._data}
            )
        )
        return f"{self.__class__.__qualname__}(\n  {args}\n)"

    def generate(self, shard_id: str) -> JsonShard[T]:
        data = self._data.generate()
        return self._generate(data=data, shard_id=shard_id)

    @abstractmethod
    def _generate(self, data: T, shard_id: str) -> JsonShard[T]:
        r"""Generate a shard based on the data and shard ID.

        Args:
            data: The data to save in the shard.
            shard_id: The shard IDI.

        Returns:
            The generated shard.
        """
